import type { AtRule, ChildNode, Container, Node } from "postcss";
import type { Screen } from "tailwindcss/types/config";

export function isCSSVariable(value: string) {
  return /^var\((--.+?)\)$/.test(value.trim());
}

export function isAtRuleNode(node: Node | undefined): node is AtRule {
  return !!(node?.type === "atrule");
}

export function isChildNode(node: Node | undefined): node is ChildNode {
  return !!(
    node &&
    (node.type === "atrule" ||
      node.type === "rule" ||
      node.type === "decl" ||
      node.type === "comment")
  );
}

export function buildMediaQueryByScreen(screens: string | Screen | Screen[]) {
  if (typeof screens === "string") {
    return `(min-width: ${screens})`;
  }

  screens = Array.isArray(screens) ? screens : [screens];

  return screens
    .map(screen => {
      if ("raw" in screen && screen["raw"]) {
        return screen["raw"];
      }

      const conditions: string[] = [];
      if ("min" in screen && screen["min"]) {
        conditions.push(`(min-width: ${screen["min"]})`);
      }

      if ("max" in screen && screen["max"]) {
        conditions.push(`(max-width: ${screen["max"]})`);
      }

      if (conditions.length) {
        return conditions.join(" and ");
      }

      return null;
    })
    .filter(Boolean)
    .join(", ");
}

export function detectIndent(root: Container<any>) {
  if (root.raws.indent) return root.raws.indent;

  let detectedIndent = "    ";
  root.walk(node => {
    const p = node.parent;
    if (
      p &&
      p !== root &&
      p.parent &&
      p.parent === root &&
      node.raws.before !== undefined
    ) {
      const parts = node.raws.before.split("\n");
      detectedIndent = parts.at(-1)!;
      detectedIndent = detectedIndent.replace(/\S/g, "");
      return false;
    }
  });

  return detectedIndent;
}

export function flattenObject(
  object: Record<string | number | symbol, any>,
  separator = "-"
) {
  const flatObject: Record<string, any> = {};

  for (const [prop, value] of Object.entries(object)) {
    if (typeof value === "object" && value !== null) {
      const nestedFlatObject = flattenObject(value, separator);

      for (const [nestedProp, nestedValue] of Object.entries(nestedFlatObject)) {
        flatObject[prop + separator + nestedProp] = nestedValue;
      }
    } else {
      flatObject[prop] = value;
    }
  }
  return flatObject;
}

export function removeUnnecessarySpaces(string: string) {
  return string.replace(/(\s+)?([,:;])(\s+)?/gm, "$2");
}
