import { promises as fs } from "node:fs";
import { resolve } from "node:path";
import { type BuildOptions, type OutputFile, build, transformSync } from "esbuild";
import { afterEach, beforeEach, describe, expect, it } from "vitest";
import dedent from "dedent";
import { name } from "../package.json" with { type: "json" };
import {
  type TailwindPluginOptions,
  babelPlugin,
  createPostCSS,
  getClassName,
  getTailwindPlugins,
} from "./index";

const folder = resolve(import.meta.dirname, "temp");

describe("babel-tailwind", () => {
  beforeEach(async () => {
    await fs.mkdir(folder, { recursive: true });
  });

  afterEach(async () => {
    await fs.rm(folder, { recursive: true, force: true });
  });

  it("supports ESBuild", async () => {
    const outputFiles = await compileESBuild({
      clsx: "emotion",
      javascript: /* tsx */ `
        export function Hello() {
          return (
            <div css="text-center">
              Hello, world!
            </div>
          );
        }
      `,
    });
    expect(outputFiles).toHaveLength(2);

    const js = findByExt(outputFiles, ".js");
    const css = findByExt(outputFiles, ".css");

    const clsName = getClassName("text-center");
    expect(js.text).toContain(`className: "${clsName}"`);
    expect(css.text).toMatch(`.${clsName} {\n  text-align: center;\n}`);
  });

  it("does not remove the attribute if `preserveAttribute` is true", async () => {
    const outputFiles = await compileESBuild({
      clsx: "emotion",
      jsxAttributeAction: "preserve",
      javascript: /* tsx */ `
        export function Hello() {
          return (
            <div css="text-center">
              Hello, world!
            </div>
          );
        }
      `,
    });
    expect(outputFiles).toHaveLength(2);

    const js = findByExt(outputFiles, ".js");
    expect(js.text).toContain(`css: "text-center"`);
  });

  it("reports errors with correct position", async () => {
    try {
      await compileESBuild({
        clsx: "emotion",
        jsxAttributeAction: "preserve",
        esbuild: {
          logLevel: "silent",
        },
        javascript: /* tsx */ `
          export function Hello() {
            return (
              <div css="text-center2 m-0">
                Hello, world!
              </div>
            );
          }
        `,
      });
      throw new Error("Expected an error");
    } catch (e) {
      expect(e.errors).toHaveLength(1);
      const [error] = e.errors;
      expect(error.location).toMatchObject({
        column: 14,
        length: 12,
        line: 3,
        lineText: '    <div css="text-center2 m-0">',
      });
    }
  });

  it("supports custom jsxAttributeName", async () => {
    const outputFiles = await compileESBuild({
      clsx: "emotion",
      jsxAttributeName: "tw",
      javascript: /* tsx */ `
        export function Hello() {
          return (
            <div tw="text-center">
              Hello, world!
            </div>
          );
        }
      `,
    });
    expect(outputFiles).toHaveLength(2);

    const js = findByExt(outputFiles, ".js");
    const css = findByExt(outputFiles, ".css");

    const clsName = getClassName("text-center");
    expect(js.text).toContain(`className: "${clsName}"`);
    expect(css.text).toMatch(`.${clsName} {\n  text-align: center;\n}`);
  });

  it("supports importing tailwind/base", async () => {
    const postcss = createPostCSS({ tailwindConfig: {} });
    const base = (await postcss("@tailwind base;")).css;
    const outputFiles = await compileESBuild({
      clsx: "emotion",
      javascript: /* tsx */ `
        import "${name}/base";
      `,
    });
    expect(outputFiles).toHaveLength(2);

    const js = findByExt(outputFiles, ".js");
    const css = findByExt(outputFiles, ".css");

    // expect(js.text).toContain(`import "./base.css";`);
    expect(js.text).toBe("");

    expect(minCSS(css.text)).toContain(minCSS(base));
  });
});

async function write(path: string, content: string) {
  const resolved = resolve(folder, path);
  await fs.writeFile(resolved, content);
  return resolved;
}

const minCSS = (text: string) =>
  transformSync(text, { minify: true, loader: "css" }).code;

const findByExt = (outputFiles: OutputFile[], ext: string) =>
  outputFiles.find(file => file.path.endsWith(ext))!;

async function compileESBuild({
  javascript,
  esbuild,
  ...options
}: TailwindPluginOptions & {
  esbuild?: BuildOptions;
  javascript: string;
}) {
  const tailwind = getTailwindPlugins({
    tailwindConfig: {},
    ...options,
  });
  const result = await build({
    bundle: true,
    write: false,
    external: ["react/jsx-runtime"],
    outdir: "dist",
    format: "esm",
    entryPoints: [await write("index.tsx", dedent(javascript))],
    plugins: [babelPlugin({ plugins: [tailwind.babel] }), tailwind.esbuild],
    ...esbuild,
  });

  const { errors, warnings, outputFiles } = result;
  expect(errors).toHaveLength(0);
  expect(warnings).toHaveLength(0);

  return outputFiles!;
}
