import hash from "@emotion/hash";
import type { Config } from "tailwindcss";
import type postcss from "postcss";
import { memoize } from "lodash";
import { babelTailwind } from "./babel-tailwind";
import { esbuildPlugin } from "./esbuild-postcss";
import { vitePlugin } from "./vite-plugin";
import { type StyleMap, createPostCSS, type tailwindDirectives } from "./shared";

export { babelPlugin } from "./esbuild-babel";
export { createPostCSS } from "./shared";

type GetClassName = (className: string) => string;

/**
 * Tagged template macro function for Tailwind classes
 * @example "tw" => tw`p-2 text-center`
 */
export type TaggedTailwindFunction = (strings: TemplateStringsArray) => string;

export interface TailwindPluginOptions {
  /**
   * Tailwind CSS configuration
   */
  tailwindConfig?: Omit<Config, "content">;

  /**
   * Directives to prefix to all Tailwind stylesheets
   */
  directives?: (typeof tailwindDirectives)[number][] | "all";

  /**
   * Additional PostCSS plugins (optional)
   */
  postCSSPlugins?: postcss.AcceptedPlugin[];

  /**
   * Attribute to use for tailwind classes in JSX
   * @default "css"
   */
  jsxAttributeName?: string;

  /**
   * What to do with the original attribute after processing
   * @default "delete"
   */
  jsxAttributeAction?: "delete" | "preserve" | ["rename", string];

  /**
   * Tagged template macro function to use for Tailwind classes
   * @default "tw"
   * @example
   * declare const tw: TaggedTailwindFunction;
   * "tw" => tw`p-2 text-center`
   */
  taggedTemplateName?: string;

  /**
   * The prefix to use for the generated class names.
   * @default className => `tw-${hash(className)}`
   */
  getClassName?: GetClassName;

  /**
   * Preferred library for classnames
   */
  clsx: "clsx" | "classnames" | "emotion";
}

/**
 * Hashes and prefixes a string of Tailwind class names.
 * @example getClassName("p-2 text-center") // "tw-1r6fxxz"
 */
export const getClassName: GetClassName = cls => "tw-" + hash(cls);

/**
 * Main entry. Returns the plugins and utilities for processing Tailwind
 * classNames in JS.
 *
 * @example
 * import { build } from "esbuild";
 * import getTailwindPlugins, { babelPlugin } from "babel-tailwind";
 *
 * const tailwind = getTailwindPlugins(options);
 * const result = await build({
 *   plugins: [
 *     babelPlugin({ getPlugins: () => [tailwind.babel] }),
 *     tailwind.esbuild,
 *   ],
 * });
 */
export function getTailwindPlugins(options: TailwindPluginOptions) {
  const styleMap: StyleMap = new Map();
  const compile = memoize(createPostCSS(options));

  return {
    compile,
    babel: babelTailwind(styleMap, options),
    esbuild: esbuildPlugin(styleMap, compile),
    vite: vitePlugin(styleMap, compile),
    styleMap,
    [Symbol.dispose]() {
      styleMap.clear();
    },
  };
}
